class CeiCrawlerUnableToLoginException(Exception):
    pass


class CeiCrawlerBlankCredentialsException(Exception):
    DEFAULT_MESSAGE = "Username ou password não podem estar em branco!"

    def __init__(self, *args: object) -> None:
        if args:
            super().__init__(*args)
        else:
            super().__init__(self.DEFAULT_MESSAGE)



import asyncio
from datetime import date as date_typing
from functools import cached_property
from typing import Any, Dict, List, Optional, Type, Union

from aiohttp import ClientSession
from aiohttp.connector import TCPConnector

from .client import CeiClient
from .models import Broker, BrokerAccount, AssetExtract, PassiveIncome
from .responses import (
    BrokerAccountAssetExtractResponse,
    BrokerAccountResponse,
    BrokerAssetsResponse,
    BrokerPassiveIncomesResponse,
    PassiveIncomesResponse,
)


class BaseCrawler:
    def __init__(
        self,
        connector: TCPConnector,
        username: str,
        password: str,
    ) -> None:
        self.connector = connector
        self.username = username
        self.password = password

    @cached_property
    def _session(self) -> ClientSession:
        return ClientSession(connector=self.connector, connector_owner=False)

    @cached_property
    def _http_client(self) -> CeiClient:
        return CeiClient(
            username=self.username,
            password=self.password,
            session=self._session,
            base_url=self.get_base_url(),
        )

    def get_base_url(self) -> str:
        raise NotImplementedError()

    def get_brokers_response_class(self) -> Any:
        raise NotImplementedError()

    def _get_broker_accounts_payload(
        self, broker: Broker
    ) -> Dict[str, Union[str, bool]]:
        raise NotImplementedError()

    async def session_close(self) -> None:
        await self._session.close()

    async def connection_close(self) -> None:
        await self.connector.close()

    async def get_brokers(self) -> List[Broker]:
        response = await self._http_client.get_brokers()
        ResponseClass = (  # pylint: disable=invalid-name
            self.get_brokers_response_class()
        )

        return await ResponseClass(response).data()

    async def get_broker_accounts(self, broker: Broker) -> Broker:
        response = await self._http_client.get_broker_accounts(
            self._get_broker_accounts_payload(broker=broker)
        )
        return await BrokerAccountResponse(response=response, broker=broker).data()

    async def get_brokers_with_accounts(self) -> List[Broker]:
        brokers = await self.get_brokers()
        tasks = [
            asyncio.create_task(self.get_broker_accounts(broker)) for broker in brokers
        ]

        return await asyncio.gather(*tasks)


class AssetsCrawler(BaseCrawler):
    def get_base_url(self) -> str:
        return "https://ceiapp.b3.com.br/CEI_Responsivo/negociacao-de-ativos.aspx"

    def get_brokers_response_class(self) -> Type[BrokerAssetsResponse]:
        return BrokerAssetsResponse

    def _get_broker_accounts_payload(
        self, broker: Broker
    ) -> Dict[str, Union[str, bool]]:
        default_account = "0"

        return {
            "ctl00$ContentPlaceHolder1$ToolkitScriptManager1": (
                "ctl00$ContentPlaceHolder1$updFiltro|ctl00$ContentPlaceHolder1"
                "$ddlAgentes"
            ),
            "__EVENTTARGET": "ctl00$ContentPlaceHolder1$ddlAgentes",
            "__VIEWSTATE": (
                "/wEPDwUKLTI2MTA0ODczNg9kFgJmD2QWAgIDD2QWCAIBDw8WAh4EVGV4dAUcR"
                "0lPVkFOTkkgQ09STkFDSElOSSBEQSBTSUxWQWRkAgMPZBYCAgEPFgIeB1Zpc2"
                "libGVoZAIFD2QWBAIBDw8WAh8ABRZOZWdvY2lhw6fDo28gZGUgYXRpdm9zZGQ"
                "CAw9kFgICAw8PFgIfAAUzIC8gRXh0cmF0b3MgZSBJbmZvcm1hdGl2b3MgLyBO"
                "ZWdvY2lhw6fDo28gZGUgYXRpdm9zZGQCBw9kFgICBQ9kFgJmD2QWCAIDDxBkE"
                "BUFCVNlbGVjaW9uZR85MCAtIEVBU1lOVkVTVCAtIFRJVFVMTyBDViBTLkEuHj"
                "MgLSBYUCBJTlZFU1RJTUVOVE9TIENDVFZNIFMvQSMzODYgLSBSSUNPIElOVkV"
                "TVElNRU5UT1MgLSBHUlVQTyBYUCAzMDggLSBDTEVBUiBDT1JSRVRPUkEgLSBH"
                "UlVQTyBYUBUFAi0xAjkwATMDMzg2AzMwOBQrAwVnZ2dnZxYBZmQCBQ8QDxYCH"
                "gdFbmFibGVkaGQQFQEMU2VsZWNpb25lLi4uFQEBMBQrAwFnFgFmZAIHDw8WAh"
                "8BZ2QWBAIFDw8WAh8ABQoxNS8wMy8yMDE5ZGQCBw8PFgIfAAUKMDQvMDkvMjA"
                "yMGRkAgkPZBYIAgEPDxYCHwAFCjE1LzAzLzIwMTlkZAIDDw8WAh8ABQowNC8"
                "wOS8yMDIwZGQCBQ8PFgIfAAUKMTUvMDMvMjAxOWRkAgcPDxYCHwAFCjA0LzA5"
                "LzIwMjBkZGRxwKn9JGbY4ybO1hczEjDtoMNWHA=="
            ),
            "__VIEWSTATEGENERATOR": "B345DEBA",
            "__EVENTVALIDATION": (
                "/wEdAA1HJ2+J0WZq5G92xJM7n1xbkdMq+p9EbLSDyFDpFsJg30jWLhbIoNprA"
                "gvONYSktR9kUdPnKtUYGLSZMc/L6M60Vx7WZDZAC7VL6Ff5PQQ8Ml0aq/zmKf"
                "YeYoLRE68kUnN+T4cvEJcdEKVEObT+JUFSdPwdZycbAgmz2rhziqC/I4jipT0"
                "jOBb54glnPZxI9VIJZJh1VhVVa0UQ15FhckhgajRdmhEnQj18gON7y5L4CwGM"
                "p96e++H/oSCxgpNo73nsffQOO+5kepETTltW2pZIXYv2PXYLNUXZnEOvnHkp6"
                "xZ0N94="
            ),
            "ctl00$ContentPlaceHolder1$ddlAgentes": broker.value,
            "ctl00$ContentPlaceHolder1$ddlContas": default_account,
            "ctl00$ContentPlaceHolder1$txtDataDeBolsa": broker.parse_extra_data.start_date,
            "ctl00$ContentPlaceHolder1$txtDataAteBolsa": broker.parse_extra_data.end_date,
            "__ASYNCPOST": True,
        }

    async def get_broker_account_portfolio_assets_extract(
        self,
        broker: Broker,
        account: BrokerAccount,
        start_date: Optional[date_typing] = None,
        end_date: Optional[date_typing] = None,
        as_dict: bool = False,
    ) -> Union[List[AssetExtract], List[Dict[str, Any]]]:
        response = await self._http_client.get_broker_account_portfolio_assets_extract(
            broker=broker,
            account=account,
            start_date=start_date,
            end_date=end_date,
        )
        return await BrokerAccountAssetExtractResponse(
            response=response, broker_value=broker.value
        ).data(as_dict=as_dict)

    async def get_brokers_account_portfolio_assets_extract(
        self,
        brokers: List[Broker],
        start_date: Optional[date_typing] = None,
        end_date: Optional[date_typing] = None,
        as_dict: bool = False,
    ) -> Union[List[AssetExtract], List[Dict[str, Any]]]:
        tasks = [
            asyncio.create_task(
                self.get_broker_account_portfolio_assets_extract(
                    broker=broker,
                    account=broker.accounts[0],
                    start_date=start_date,
                    end_date=end_date,
                    as_dict=as_dict,
                )
            )
            for broker in brokers
            if broker.accounts
        ]

        results = await asyncio.gather(*tasks)
        return [result for subresult in results for result in subresult]


class PassiveIncomesCrawler(BaseCrawler):
    def get_base_url(self) -> str:
        return "https://ceiapp.b3.com.br/CEI_Responsivo/ConsultarProventos.aspx"

    def get_brokers_response_class(self) -> Type[BrokerPassiveIncomesResponse]:
        return BrokerPassiveIncomesResponse

    def _get_broker_accounts_payload(
        self, broker: Broker
    ) -> Dict[str, Union[str, bool]]:
        default_account = "0"

        return {
            "ctl00$ContentPlaceHolder1$ToolkitScriptManager1": (
                "ctl00$ContentPlaceHolder1$updFiltro|ctl00$ContentPlaceHolder1"
                "$ddlAgentes"
            ),
            "__EVENTTARGET": "ctl00$ContentPlaceHolder1$ddlAgentes",
            "__VIEWSTATE": (
                "/wEPDwULLTEwMDg3NTU3MTAPZBYCZg9kFgICAw9kFggCAQ8PFgIeBFRleHQFF1dBR05"
                "FUiBWSUVMTU8gREUgQ0FNUE9TZGQCAw9kFgICAQ8WAh4HVmlzaWJsZWhkAgUPZBYEAgE"
                "PDxYCHwAFCVByb3ZlbnRvc2RkAgMPZBYCAgMPDxYCHwAFHCAvIEludmVzdGltZW50b3M"
                "gLyBQcm92ZW50b3NkZAIHD2QWAgIFD2QWAmYPZBYMAgEPZBYCAgEPEGQQFQMFVG9kb3M"
                "WMTk4MiAtIE1PREFMIERUVk0gTFREQSMzODYgLSBSSUNPIElOVkVTVElNRU5UT1MgLSB"
                "HUlVQTyBYUBUDATAEMTk4MgMzODYUKwMDZ2dnFgFmZAIDD2QWAgIBDxAPFgIeB0VuYWJ"
                "sZWRoZBAVAwVUb2RvcwYzNDUwMzAHMTg1OTAzNhUDATAGMzQ1MDMwBzE4NTkwMzYUKwM"
                "DZ2dnFgFmZAIFDxYCHwFoFgICAQ8QZGQWAGQCBw9kFgYCAQ8PFgQeCENzc0NsYXNzBQp"
                "kYXRlcGlja2VyHgRfIVNCAgJkZAIDDw8WAh8ABQoxNC8wNi8yMDIxZGQCBQ8PFgIfAAU"
                "KMTcvMDYvMjAyMWRkAgsPFgIfAWcWAgIBDw8WAh8ABSFBdHVhbGl6YWRvIGVtIDE3LzA"
                "2LzIwMjEgYXMgMjI6MDlkZAINDxYCHgtfIUl0ZW1Db3VudAIBFgJmD2QWBAIBDw8WAh8"
                "ABSMzODYgLSBSSUNPIElOVkVTVElNRU5UT1MgLSBHUlVQTyBYUGRkAgMPFgIfBQIBFgJ"
                "mD2QWCgIBDw8WAh8ABRNDb250YSBuwrogMTg1OTAzNi05ZGQCAw9kFgICAQ8WAh8FAgI"
                "WBgIBD2QWAmYPFQkMRU5FUkdJQVMgQlIgCk9OICAgICAgTk0MRU5CUjMgICAgICAgCjI"
                "zLzA2LzIwMjEcSlVST1MgU09CUkUgQ0FQSVRBTCBQUsOTUFJJTwU0LDAwMAExBDEsMDg"
                "EMCw5MmQCAg9kFgJmDxUJDEZJSSBDQVBJIFNFQwpDSSAgICAgICAgDENQVFMxMSAgICA"
                "gIAoxOC8wNi8yMDIxClJFTkRJTUVOVE8FMSwwMDABMQQxLDAwBDEsMDBkAgMPZBYEAgE"
                "PDxYCHwAFBDIsMDhkZAIDDw8WAh8ABQQxLDkyZGQCBQ9kFgICAQ8WAh8FAhEWJAIBD2Q"
                "WAmYPFQkMRklJIEJDIEZGSUkgCkNJICAgICAgICAMQkNGRjExICAgICAgCjE1LzA2LzIw"
                "MjEKUkVORElNRU5UTwUzLDAwMAExBDEsNTAEMSw1MGQCAg9kFgJmDxUJDEZJSSBCRUVTI"
                "ENSSQpDSSAgICAgICAgDEJDUkkxMSAgICAgIAoxNS8wNi8yMDIxClJFTkRJTUVOVE8GMT"
                "AsMDAwATEFMTUsOTAFMTUsOTBkAgMPZBYCZg8VCQxGSUkgQ1NIRyBMT0cKQ0kgICAgICAg"
                "IAxIR0xHMTEgICAgICAKMTUvMDYvMjAyMQpSRU5ESU1FTlRPBjExLDAwMAExBTExLDAwBT"
                "ExLDAwZAIED2QWAmYPFQkMRklJIENTSEcgVVJCCkNJICAgICAgICAMSEdSVTExICAgICAg"
                "CjE1LzA2LzIwMjEKUkVORElNRU5UTwYxMCwwMDABMQQ3LDAwBDcsMDBkAgUPZBYCZg8VCQ"
                "xGSUkgREVWQU5UICAKQ0kgICAgICAgIAxERVZBMTEgICAgICAKMTUvMDYvMjAyMQpSRU5E"
                "SU1FTlRPBjIwLDAwMAExBTIyLDAwBTIyLDAwZAIGD2QWAmYPFQkMRklJIERFVkFOVCAgCl"
                "JFQyAgICAgICAMREVWQTEzICAgICAgCjE1LzA2LzIwMjEKUkVORElNRU5UTwYxMSwwMDAB"
                "MQQxLDQzBDEsNDNkAgcPZBYCZg8VCQxGSUkgRkFUT1IgVkUKQ0kgICAgICAgIAxWUlRBMT"
                "EgICAgICAKMTUvMDYvMjAyMQpSRU5ESU1FTlRPBTEsMDAwATEEMSwwNAQxLDA0ZAIID2QW"
                "AmYPFQkMRklJIEhHIFJFQUwgCkNJICAgICAgICAMSEdSRTExICAgICAgCjE1LzA2LzIwMj"
                "EKUkVORElNRU5UTwU2LDAwMAExBDQsMTQENCwxNGQCCQ9kFgJmDxUJDEZJSSBJUklESVVN"
                "IApDSSAgICAgICAgDElSRE0xMSAgICAgIAoxNy8wNi8yMDIxClJFTkRJTUVOVE8GMjgsMD"
                "AwATEFMzIsMzAFMzIsMzBkAgoPZBYCZg8VCQxGSUkgS0lORUEgICAKQ0kgICAgICAgIAxL"
                "TlJJMTEgICAgICAKMTUvMDYvMjAyMQpSRU5ESU1FTlRPBjEwLDAwMAExBDYsOTAENiw5MG"
                "QCCw9kFgJmDxUJDEZJSSBLSU5FQSBJUApDSSAgICAgICAgDEtOSVAxMSAgICAgIAoxNC8w"
                "Ni8yMDIxClJFTkRJTUVOVE8FNSwwMDABMQQ1LDY1BDUsNjVkAgwPZBYCZg8VCQxGSUkgTU"
                "FYSSBSRU4KQ0kgICAgICAgIAxNWFJGMTEgICAgICAKMTUvMDYvMjAyMQpSRU5ESU1FTlRP"
                "BzE2NSwwMDABMQUxMSw1NQUxMSw1NWQCDQ9kFgJmDxUJDEZJSSBUT1JERSBFSQpDSSAgICA"
                "gICAgDFRPUkQxMSAgICAgIAoxNS8wNi8yMDIxClJFTkRJTUVOVE8GNDcsMDAwATEEMywyOQ"
                "QzLDI5ZAIOD2QWAmYPFQkMRklJIFRPUkRFIEVJClJFQyAgICAgICAMVE9SRDEzICAgICAgC"
                "jE1LzA2LzIwMjEKUkVORElNRU5UTwYxNSwwMDABMQQwLDU1BDAsNTVkAg8PZBYCZg8VCQxG"
                "SUkgVkVSUyBDUkkKQ0kgICAgICAgIAxWU0xIMTEgICAgICAKMTUvMDYvMjAyMQpSRU5ESU1"
                "FTlRPBTUsMDAwATEEMCw4NQQwLDg1ZAIQD2QWAmYPFQkMRklJIFZJTkNJTE9HCkNJICAgIC"
                "AgICAMVklMRzExICAgICAgCjE1LzA2LzIwMjEKUkVORElNRU5UTwU0LDAwMAExBDIsMjgEM"
                "iwyOGQCEQ9kFgJmDxUJDEZJSSBYUCBMT0cgIApDSSAgICAgICAgDFhQTEcxMSAgICAgIAox"
                "NS8wNi8yMDIxClJFTkRJTUVOVE8FNiwwMDABMQQzLDY2BDMsNjZkAhIPZBYEAgEPDxYCHwA"
                "FBjEzMSwwNGRkAgMPDxYCHwAFBjEzMSwwNGRkAgcPFgIfAWgWAgIBDxYCHwUC/////w9kAg"
                "kPZBYCAgEPFgIfBQIBFgQCAQ9kFgJmDxULDEZJSSBUT1JERSBFSQpSRUMgICAgICAgDFRPU"
                "kQxMyAgICAgIAtBVFVBTElaQUNBTwoxNy8wNi8yMDIxBjE1LDAwMAExDFRPUkQxMSAgICAg"
                "IAUxNSwwMAYxMDAsMDAEMCwwMGQCAg9kFgICAQ8PFgIfAAUEMCwwMGRkZNmTebZZFDui/ij"
                "Eax1lX4ugM/e1"
            ),
            "__VIEWSTATEGENERATOR": "60D025E9",
            "__EVENTVALIDATION": (
                "/wEdAAz3Yx2ES4cT5dCIgXUVBTgKSNYuFsig2msCC841hKS1H1GNyhN3N1osq"
                "BqqgaNItpdZDO+9C9gz5Aw/AZYSIzXhfk+HLxCXHRClRDm0/iVBUojipT0jOB"
                "b54glnPZxI9VIJZJh1VhVVa0UQ15FhckhgY2JNro3bvLeSMJrpn9aSiHkRxwu"
                "EE8o8LwrE2E6mACtDmhxe4pZ6Ibt8oBWfF3Ts7H30DjvuZHqRE05bVtqWSPOO"
                "FWTf/Hn+MVE/ujXYFJTWLVokCvgbtFxjrfYgYXuQJX3w9w=="
            ),
            "ctl00$ContentPlaceHolder1$ddlAgentes": broker.value,
            "ctl00$ContentPlaceHolder1$ddlContas": default_account,
            "ctl00$ContentPlaceHolder1$txtData": broker.parse_extra_data.end_date,
            "__ASYNCPOST": True,
        }

    async def get_passive_incomes_extract(
        self, date: Optional[date_typing] = None, as_dict: bool = False
    ) -> Union[List[PassiveIncome], List[Dict[str, Any]]]:
        brokers = await self.get_brokers_with_accounts()
        response = await self._http_client.get_passive_incomes_extract(
            broker=brokers[0],  # brokers[0] is the broker with all accounts
            date=date,
        )

        return await PassiveIncomesResponse(response=response).data(as_dict=as_dict)



from datetime import date as date_typing
from functools import cached_property
from typing import Any, Dict, List, Optional, Union

from aiohttp import TCPConnector

from .crawlers import AssetsCrawler, PassiveIncomesCrawler
from .exceptions import CeiCrawlerBlankCredentialsException
from .models import Broker, AssetExtract, PassiveIncome

BLANK_VALUES = (None, "")


class CeiCrawler:
    def __init__(self, username: str, password: str) -> None:
        if username in BLANK_VALUES or password in BLANK_VALUES:
            raise CeiCrawlerBlankCredentialsException()

        self._connector = TCPConnector(limit=30, ssl=False)
        self.username = username
        self.password = password

    @cached_property
    def _assets_crawler(self) -> AssetsCrawler:
        return AssetsCrawler(
            connector=self._connector,
            username=self.username,
            password=self.password,
        )

    @cached_property
    def _passive_incomes_crawler(self) -> PassiveIncomesCrawler:
        return PassiveIncomesCrawler(
            connector=self._connector,
            username=self.username,
            password=self.password,
        )

    async def close(self) -> None:
        await self._assets_crawler.session_close()
        await self._passive_incomes_crawler.session_close()
        await self._connector.close()

    async def get_brokers(self) -> List[Broker]:
        return await self._assets_crawler.get_brokers_with_accounts()

    async def get_assets_extract(
        self,
        brokers: Optional[List[Broker]] = None,
        start_date: Optional[date_typing] = None,
        end_date: Optional[date_typing] = None,
        as_dict: bool = False,
    ) -> Union[List[AssetExtract], List[Dict[str, Any]]]:
        return await self._assets_crawler.get_brokers_account_portfolio_assets_extract(
            brokers=brokers if brokers is not None else await self.get_brokers(),
            start_date=start_date,
            end_date=end_date,
            as_dict=as_dict,
        )

    async def get_passive_incomes_extract(
        self, date: Optional[date_typing] = None, as_dict: bool = False
    ) -> Union[List[PassiveIncome], List[Dict[str, Any]]]:
        return await self._passive_incomes_crawler.get_passive_incomes_extract(
            date=date, as_dict=as_dict
        )



import pytest

from ..constants import (
    ASSET_ACTION_TYPE_MAPPER,
    ASSET_MARKET_TYPE_MAPPER,
    AssetExtractAction,
    AssetExtractMarketType,
    PassiveIncomeType,
    PASSIVE_INCOME_TYPE_MAPPER,
    PassiveIncomeEventType,
    PASSIVE_INCOME_EVENT_TYPE_MAPPER,
)


class TestConstantsMapper:
    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "key, expected_value",
        [
            ("C", AssetExtractAction.BUY.value),
            ("V", AssetExtractAction.SELL.value),
        ],
    )
    async def test_asset_action_type_mapper(self, key, expected_value):
        assert ASSET_ACTION_TYPE_MAPPER[key] == expected_value

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "key, expected_value",
        [
            ("Merc. Fracionário", AssetExtractMarketType.FRACTIONAL.value),
            ("Mercado a Vista", AssetExtractMarketType.UNIT.value),
        ],
    )
    async def test_asset_market_type_mapper(self, key, expected_value):
        assert ASSET_MARKET_TYPE_MAPPER[key] == expected_value

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "key, expected_value",
        [
            ("Provisionado", PassiveIncomeEventType.FUTURE.value),
            ("Creditado", PassiveIncomeEventType.PAST.value),
        ],
    )
    async def test_passive_income_event_type_mapper(self, key, expected_value):
        assert PASSIVE_INCOME_EVENT_TYPE_MAPPER[key] == expected_value

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "key, expected_value",
        [
            ("DIVIDENDO", PassiveIncomeType.DIVIDEND.value),
            ("JUROS SOBRE CAPITAL PRÓPRIO", PassiveIncomeType.JCP.value),
            ("RENDIMENTO", PassiveIncomeType.YIELD.value),
        ],
    )
    async def test_passive_income_type_mapper(self, key, expected_value):
        assert PASSIVE_INCOME_TYPE_MAPPER[key] == expected_value



from datetime import date
from decimal import Decimal

import pytest

from ..constants import (
    AssetExtractAction,
    AssetExtractMarketType,
    PassiveIncomeEventType,
    PassiveIncomeType,
)
from ..models import AssetExtract, PassiveIncome, _format_string_to_decimal


class TestAssetExtract:
    @pytest.mark.asyncio
    async def test_create_asset_from_response_data(self):
        action = "C"
        asset_specification = "AZUL        PN      N2"
        market_type = "Merc. Fracionário"
        operation_date = "05/06/2020"
        quotation_factor = "1"
        raw_negotiation_code = "AZUL4F"
        total_price = "42,18"
        unit_amount = "2"
        unit_price = "21,09"

        expected_obj = AssetExtract(
            action=AssetExtractAction.BUY.value,
            asset_specification="AZUL PN N2",
            market_type=AssetExtractMarketType.FRACTIONAL.value,
            operation_date=date(2020, 6, 5),
            quotation_factor=1,
            raw_negotiation_code=raw_negotiation_code,
            total_price=Decimal("42.18"),
            unit_amount=2,
            unit_price=Decimal("21.09"),
        )

        broker_asset_extract = AssetExtract.create_from_response_fields(
            action=action,
            asset_specification=asset_specification,
            market_type=market_type,
            operation_date=operation_date,
            quotation_factor=quotation_factor,
            raw_negotiation_code=raw_negotiation_code,
            total_price=total_price,
            unit_amount=unit_amount,
            unit_price=unit_price,
        )

        assert await broker_asset_extract == expected_obj

    @pytest.mark.asyncio
    async def test_format_string_to_decimal_with_high_value(self):
        expected_value = Decimal("1485.12")
        unit_price = "1.485.12"

        value = _format_string_to_decimal(unit_price)

        assert await value == expected_value

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "expected_value, unit_amount",
        [(1, "1"), (1100, "1.100"), (1000000, "1.000.000")],
    )
    async def test_should_format_asset_string_to_int(self, expected_value, unit_amount):
        # GIVEN

        # WHEN
        value = AssetExtract._format_string_to_int(unit_amount)

        # THEN
        assert await value == expected_value

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "expected_value, unit_amount",
        [
            (1, "1,00"),
            (100, "100,00"),
            (1100, "1.100,00"),
            (1000000, "1.000.000,00"),
        ],
    )
    async def test_should_format_passive_income_string_to_int(
        self, expected_value, unit_amount
    ):
        # GIVEN

        # WHEN
        value = PassiveIncome._format_string_to_int(unit_amount)

        # THEN
        assert await value == expected_value

    @pytest.mark.asyncio
    async def test_create_passive_income_from_response_data(self):
        # GIVEN
        raw_negotiation_name = "ALUPAR"
        asset_specification = "UNT      N2"
        raw_negotiation_code = "ALUP11"
        operation_date = "05/06/2020"
        income_type = "DIVIDENDO"
        unit_amount = "2"
        quotation_factor = "1"
        gross_value = "35,70"
        net_value = "35,70"
        event_type = "Provisionado"

        expected_obj = PassiveIncome(
            raw_negotiation_name=raw_negotiation_name,
            asset_specification="UNT N2",
            raw_negotiation_code=raw_negotiation_code,
            operation_date=date(2020, 6, 5),
            income_type=PassiveIncomeType.DIVIDEND.value,
            unit_amount=2,
            quotation_factor=1,
            gross_value=Decimal("35.70"),
            net_value=Decimal("35.70"),
            event_type=PassiveIncomeEventType.FUTURE.value,
        )

        # WHEN
        passive_income = PassiveIncome.create_from_response_fields(
            raw_negotiation_name=raw_negotiation_name,
            asset_specification=asset_specification,
            raw_negotiation_code=raw_negotiation_code,
            operation_date=operation_date,
            event_type=event_type,
            unit_amount=unit_amount,
            quotation_factor=quotation_factor,
            gross_value=gross_value,
            net_value=net_value,
            income_type=income_type,
        )

        # THEN
        assert await passive_income == expected_obj



from enum import Enum, unique


@unique
class AssetExtractAction(Enum):
    BUY = "buy"
    SELL = "sell"


@unique
class AssetExtractMarketType(Enum):
    FRACTIONAL = "fractional_share"
    UNIT = "unit"
    OPTIONS = "options"


ASSET_ACTION_TYPE_MAPPER = {
    "C": AssetExtractAction.BUY.value,
    "V": AssetExtractAction.SELL.value,
}

ASSET_MARKET_TYPE_MAPPER = {
    "Merc. Fracionário": AssetExtractMarketType.FRACTIONAL.value,
    "Mercado a Vista": AssetExtractMarketType.UNIT.value,
    "Opção de Compra": AssetExtractMarketType.OPTIONS.value,
    "Opção de Venda": AssetExtractMarketType.OPTIONS.value,
    "Exercicio de Opções": AssetExtractMarketType.OPTIONS.value,
}


@unique
class PassiveIncomeType(Enum):
    DIVIDEND = "dividend"
    JCP = "jcp"
    YIELD = "FII yield"


PASSIVE_INCOME_TYPE_MAPPER = {
    "DIVIDENDO": PassiveIncomeType.DIVIDEND.value,
    "JUROS SOBRE CAPITAL PRÓPRIO": PassiveIncomeType.JCP.value,
    "RENDIMENTO": PassiveIncomeType.YIELD.value,
}


@unique
class PassiveIncomeEventType(Enum):
    FUTURE = "provisioned"
    PAST = "credited"


PASSIVE_INCOME_EVENT_TYPE_MAPPER = {
    "Provisionado": PassiveIncomeEventType.FUTURE.value,
    "Creditado": PassiveIncomeEventType.PAST.value,
}



from __future__ import annotations

import re
from dataclasses import dataclass, field
from datetime import datetime
from decimal import Decimal
from typing import List

from .constants import (
    ASSET_ACTION_TYPE_MAPPER,
    ASSET_MARKET_TYPE_MAPPER,
    PASSIVE_INCOME_EVENT_TYPE_MAPPER,
    PASSIVE_INCOME_TYPE_MAPPER,
    AssetExtractAction,
    AssetExtractMarketType,
    PassiveIncomeEventType,
    PassiveIncomeType,
)


@dataclass
class BrokerParseExtraData:
    start_date: str
    end_date: str


@dataclass
class Broker:
    value: str
    name: str
    parse_extra_data: BrokerParseExtraData
    accounts: List = field(default_factory=lambda: [])


@dataclass
class BrokerAccountParseExtraData:
    """Essential data to do other requests using account information."""

    view_state: str
    view_state_generator: str
    event_validation: str


@dataclass
class BrokerAccount:
    id: str
    parse_extra_data: BrokerAccountParseExtraData


async def _format_string_to_decimal(value: str) -> Decimal:
    value = value.replace(",", "").replace(".", "")
    value = f"{value[:-2]}.{value[-2:]}"
    return Decimal(value)


async def _remove_extra_space(text: str) -> str:
    return re.sub(" +", " ", text)


@dataclass
class AssetExtract:
    operation_date: datetime
    action: AssetExtractAction
    market_type: AssetExtractMarketType
    raw_negotiation_code: str
    asset_specification: str
    unit_amount: int
    unit_price: Decimal
    total_price: Decimal
    quotation_factor: int

    @classmethod
    async def create_from_response_fields(
        cls,
        operation_date,
        action,
        market_type,
        raw_negotiation_code,
        asset_specification,
        unit_amount,
        unit_price,
        total_price,
        quotation_factor,
    ) -> AssetExtract:
        action = ASSET_ACTION_TYPE_MAPPER[action]
        market_type = ASSET_MARKET_TYPE_MAPPER[market_type]
        asset_specification = await _remove_extra_space(asset_specification)
        operation_date = datetime.strptime(operation_date, "%d/%m/%Y").date()
        total_price = await _format_string_to_decimal(total_price)
        unit_price = await _format_string_to_decimal(unit_price)
        unit_amount = await cls._format_string_to_int(unit_amount)
        quotation_factor = int(quotation_factor)

        return cls(
            operation_date=operation_date,
            action=action,
            market_type=market_type,
            raw_negotiation_code=raw_negotiation_code,
            asset_specification=asset_specification,
            unit_amount=unit_amount,
            unit_price=unit_price,
            total_price=total_price,
            quotation_factor=quotation_factor,
        )

    @staticmethod
    async def _format_string_to_int(value: str) -> int:
        return int(value.replace(".", ""))


@dataclass
class PassiveIncome:
    raw_negotiation_name: str
    asset_specification: str
    raw_negotiation_code: str
    operation_date: datetime
    event_type: PassiveIncomeEventType
    unit_amount: int
    quotation_factor: int
    gross_value: Decimal
    net_value: Decimal
    income_type: PassiveIncomeType

    @classmethod
    async def create_from_response_fields(
        cls,
        raw_negotiation_name,
        asset_specification,
        raw_negotiation_code,
        operation_date,
        event_type,
        unit_amount,
        quotation_factor,
        gross_value,
        net_value,
        income_type,
    ) -> PassiveIncome:
        asset_specification = await _remove_extra_space(asset_specification)
        operation_date = datetime.strptime(operation_date, "%d/%m/%Y").date()
        event_type = PASSIVE_INCOME_EVENT_TYPE_MAPPER.get(event_type)
        unit_amount = await cls._format_string_to_int(unit_amount)
        quotation_factor = int(quotation_factor)
        gross_value = await _format_string_to_decimal(gross_value)
        net_value = await _format_string_to_decimal(net_value)
        income_type = PASSIVE_INCOME_TYPE_MAPPER.get(income_type)

        return cls(
            raw_negotiation_name=raw_negotiation_name,
            asset_specification=asset_specification,
            raw_negotiation_code=raw_negotiation_code,
            operation_date=operation_date,
            event_type=event_type,
            unit_amount=unit_amount,
            quotation_factor=quotation_factor,
            gross_value=gross_value,
            net_value=net_value,
            income_type=income_type,
        )

    @staticmethod
    async def _format_string_to_int(value: str) -> int:
        return int(value.split(",")[0].replace(".", ""))
